#!/usr/bin/env python
# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import json
import logging
import struct
import time
from collections import namedtuple
from contextlib import contextmanager
from datetime import datetime
from math import ceil
from math import floor
from typing import Iterator
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple

import a_sync
import gevent
import requests
from gevent import monkey
from gevent import pool
from kazoo.client import KazooClient
from kazoo.exceptions import NoNodeError
from marathon.models.app import MarathonApp
from marathon.models.app import MarathonTask

from paasta_tools.autoscaling.forecasting import get_forecast_policy
from paasta_tools.autoscaling.utils import get_autoscaling_component
from paasta_tools.autoscaling.utils import register_autoscaling_component
from paasta_tools.bounce_lib import LockHeldException
from paasta_tools.bounce_lib import LockTimeout
from paasta_tools.bounce_lib import ZK_LOCK_CONNECT_TIMEOUT_S
from paasta_tools.long_running_service_tools import compose_autoscaling_zookeeper_root
from paasta_tools.long_running_service_tools import set_instances_for_marathon_service
from paasta_tools.long_running_service_tools import ZK_PAUSE_AUTOSCALE_PATH
from paasta_tools.marathon_tools import AutoscalingParamsDict
from paasta_tools.marathon_tools import format_job_id
from paasta_tools.marathon_tools import get_marathon_apps_with_clients
from paasta_tools.marathon_tools import get_marathon_clients
from paasta_tools.marathon_tools import get_marathon_servers
from paasta_tools.marathon_tools import is_old_task_missing_healthchecks
from paasta_tools.marathon_tools import is_task_healthy
from paasta_tools.marathon_tools import MarathonServiceConfig
from paasta_tools.marathon_tools import MESOS_TASK_SPACER
from paasta_tools.mesos.task import Task
from paasta_tools.mesos_tools import get_all_running_tasks
from paasta_tools.mesos_tools import get_cached_list_of_running_tasks_from_frameworks
from paasta_tools.paasta_service_config_loader import PaastaServiceConfigLoader
from paasta_tools.utils import _log
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_user_agent
from paasta_tools.utils import list_services
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import mean
from paasta_tools.utils import SystemPaastaConfig
from paasta_tools.utils import use_requests_cache
from paasta_tools.utils import ZookeeperPool
try:
    import yelp_meteorite
except ImportError:
    # Sorry to any non-yelpers but you won't
    # get metrics emitted as our metrics library
    # is currently not open source
    yelp_meteorite = None

ServiceAutoscalingInfo = namedtuple(
    'ServiceAutoscalingInfo', [
        'current_instances',
        'max_instances',
        'min_instances',
        'current_utilization',
        'target_instances',
    ],
)


SERVICE_METRICS_PROVIDER_KEY = 'metrics_provider'
DECISION_POLICY_KEY = 'decision_policy'

AUTOSCALING_DELAY = 300
MAX_TASK_DELTA = 0.3

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def get_service_metrics_provider(name):
    """
    Returns a service metrics provider matching the given name.
    """
    return get_autoscaling_component(name, SERVICE_METRICS_PROVIDER_KEY)


def get_decision_policy(name):
    """
    Returns a decision policy matching the given name.
    Decision policies determine the direction a service needs to be scaled in.
    Each decision policy returns one of the following values:
    -1: autoscale down
    0:  don't autoscale
    1:  autoscale up
    """
    return get_autoscaling_component(name, DECISION_POLICY_KEY)


class MetricsProviderNoDataError(ValueError):
    pass


@register_autoscaling_component('threshold', DECISION_POLICY_KEY)
def threshold_decision_policy(current_instances, error, **kwargs):
    """
    Decides to autoscale up or down by 10% if the service exceeds the upper or lower thresholds
    (see get_error_from_value() for how the thresholds are created)
    """
    autoscaling_amount = max(1, int(current_instances * 0.1))
    if error > 0:
        return autoscaling_amount
    elif abs(error) > 0:
        return -autoscaling_amount
    else:
        return 0


@register_autoscaling_component('proportional', DECISION_POLICY_KEY)
def proportional_decision_policy(
    zookeeper_path, current_instances, min_instances, max_instances, setpoint, utilization,
    num_healthy_instances, noop=False, offset=0.0, forecast_policy='current',
    good_enough_window=None, **kwargs,
):
    """Uses a simple proportional model to decide the correct number of instances to scale to, i.e. if load is 110% of
    the setpoint, scales up by 10%. Includes correction for an offset, if your containers have a baseline utilization
    independent of the number of containers.

    The model is: utilization per container = (total load)/(number of containers) + offset.

    total load and offset are measured in the same unit as your metric provider. If you're measuring CPU per container,
    offset is the baseline CPU of an idle container, and total load is the total CPU required across all containers,
    subtracting the offset for each container.

    :param offset: A float (should be between 0.0 and 1.0) representing the expected baseline load for each container.
                   e.g. if the metric you're using is CPU, then how much CPU an idle container would use.
                   This should never be more than your setpoint. (If it takes 50% cpu to run an idle container, we can't
                   get your utilization below 50% no matter how many containers we run.)
    :param forecast_policy: The method for forecasting future load values. Currently, only two forecasters exist:
                            - "current", which assumes that the load will remain the same as the current value for the
                            near future.
                            - "moving_average", which assumes that total load will remain near the average of data
                            points within a window.
    :param good_enough_window: A tuple/array of two utilization values, (low, high). If the utilization per container at
                               the forecasted total load is within this window with the current number of instances,
                               leave the number of instances alone. This can reduce churn. Setpoint should lie within
                               this window.
    """

    forecast_policy_func = get_forecast_policy(forecast_policy)

    current_load = (utilization - offset) * num_healthy_instances

    historical_load = fetch_historical_load(zk_path_prefix=zookeeper_path)
    historical_load.append((time.time(), current_load))
    save_historical_load(historical_load, zk_path_prefix=zookeeper_path)

    predicted_load = forecast_policy_func(historical_load, **kwargs)

    desired_number_instances = int(round(predicted_load / (setpoint - offset)))

    if good_enough_window:
        low, high = good_enough_window
        predicted_load_per_instance_with_current_instances = predicted_load / current_instances + offset
        if low <= predicted_load_per_instance_with_current_instances <= high:
            desired_number_instances = current_instances

    if desired_number_instances < min_instances:
        desired_number_instances = min_instances
    if desired_number_instances > max_instances:
        desired_number_instances = max_instances

    return desired_number_instances - current_instances  # The calling function wants a delta, not an absolute value.


HISTORICAL_LOAD_SERIALIZATION_FORMAT = 'dd'
SIZE_PER_HISTORICAL_LOAD_RECORD = struct.calcsize(HISTORICAL_LOAD_SERIALIZATION_FORMAT)


def zk_historical_load_path(zk_path_prefix):
    return "%s/historical_load" % zk_path_prefix


def save_historical_load(historical_load, zk_path_prefix):
    with ZookeeperPool() as zk:
        historical_load_bytes = serialize_historical_load(historical_load)
        zk.ensure_path(zk_historical_load_path(zk_path_prefix))
        zk.set(zk_historical_load_path(zk_path_prefix), historical_load_bytes)


def serialize_historical_load(historical_load):
    max_records = 1000000 // SIZE_PER_HISTORICAL_LOAD_RECORD
    historical_load = historical_load[-max_records:]
    return b''.join([struct.pack(HISTORICAL_LOAD_SERIALIZATION_FORMAT, *x) for x in historical_load])


def fetch_historical_load(zk_path_prefix):
    with ZookeeperPool() as zk:
        try:
            historical_load_bytes, _ = zk.get(zk_historical_load_path(zk_path_prefix))
            return deserialize_historical_load(historical_load_bytes)
        except NoNodeError:
            return []


def deserialize_historical_load(historical_load_bytes):
    historical_load = []

    for pos in range(0, len(historical_load_bytes), SIZE_PER_HISTORICAL_LOAD_RECORD):
        historical_load.append(
            struct.unpack(
                # unfortunately struct.unpack doesn't like kwargs.
                HISTORICAL_LOAD_SERIALIZATION_FORMAT,
                historical_load_bytes[pos:pos + SIZE_PER_HISTORICAL_LOAD_RECORD],
            ),
        )

    return historical_load


def get_json_body_from_service(host, port, endpoint, timeout=2):
    return requests.get(
        f'http://{host}:{port}/{endpoint}',
        headers={'User-Agent': get_user_agent()}, timeout=timeout,
    ).json()


def get_http_utilization_for_a_task(task, service, endpoint, json_mapper):
    """
    Gets the task utilization by fetching json from an http endpoint
    and applying a function that maps it to a utilization.

    :param task: the Marathon task to get data from
    :param service: service name
    :param endpoint: the http endpoint to get the task stats from
    :param json_mapper: a function that takes a dictionary for a task and returns that task's utilization

    :returns: the service's utilization, from 0 to 1, or None
    """
    try:
        return json_mapper(get_json_body_from_service(task.host, task.ports[0], endpoint))
    except requests.exceptions.Timeout:
        # If we time out querying an endpoint, assume the task is fully loaded
        # This won't trigger in the event of DNS error or when a request is refused
        # a requests.exception.ConnectionError is raised in those cases
        log.debug("Received a timeout when querying %s on %s:%s. Assuming the service "
                  "is at full utilization." % (service, task.host, task.ports[0]))
        return 1.0
    except Exception as e:
        log.error("Caught exception when querying {} on {}:{} : {}".format(service, task.host, task.ports[0], str(e)))


def get_http_utilization_for_all_tasks(marathon_service_config, marathon_tasks, endpoint, json_mapper):
    """
    Gets the mean utilization of a service across all of its tasks by fetching
    json from an http endpoint and applying a function that maps it to a
    utilization

    :param marathon_service_config: the MarathonServiceConfig to get data from
    :param marathon_tasks: Marathon tasks to get data from
    :param endpoint: The http endpoint to get the stats from
    :param json_mapper: A function that takes a dictionary for a task and returns that task's utilization

    :returns: the service's mean utilization, from 0 to 1
    """

    endpoint = endpoint.lstrip('/')
    utilization = []
    service = marathon_service_config.get_service()

    monkey.patch_socket()
    gevent_pool = pool.Pool(20)
    jobs = [
        gevent_pool.spawn(get_http_utilization_for_a_task, task, service, endpoint, json_mapper)
        for task in marathon_tasks
    ]
    gevent.joinall(jobs)

    for job in jobs:
        if job.value is not None:
            utilization.append(job.value)

    if not utilization:
        raise MetricsProviderNoDataError("Couldn't get any data from http endpoint {} for {}.{}".format(
            endpoint, marathon_service_config.service, marathon_service_config.instance,
        ))
    return mean(utilization)


@register_autoscaling_component('uwsgi', SERVICE_METRICS_PROVIDER_KEY)
def uwsgi_metrics_provider(marathon_service_config, marathon_tasks, endpoint='status/uwsgi', **kwargs):
    """
    Gets the mean utilization of a service across all of its tasks, where
    the utilization of a task is the percentage of non-idle workers as read
    from the UWSGI stats endpoint

    :param marathon_service_config: the MarathonServiceConfig to get data from
    :param marathon_tasks: Marathon tasks to get data from
    :param endpoint: The http endpoint to get the uwsgi stats from

    :returns: the service's mean utilization, from 0 to 1
    """

    def uwsgi_mapper(json):
        workers = json['workers']
        utilization = [1.0 if worker['status'] != 'idle' else 0.0 for worker in workers]
        return mean(utilization)

    return get_http_utilization_for_all_tasks(marathon_service_config, marathon_tasks, endpoint, uwsgi_mapper)


@register_autoscaling_component('http', SERVICE_METRICS_PROVIDER_KEY)
def http_metrics_provider(marathon_service_config, marathon_tasks, endpoint='status', **kwargs):
    """
    Gets the mean utilization of a service across all of its tasks, where
    the utilization of a task is read from a HTTP endpoint on the host. The
    HTTP endpoint must return JSON with a 'utilization' key with a value from 0
    to 1.

    :param marathon_service_config: the MarathonServiceConfig to get data from
    :param marathon_tasks: Marathon tasks to get data from
    :param endpoint: The http endpoint to get the task utilization from

    :returns: the service's mean utilization, from 0 to 1
    """

    def utilization_mapper(json):
        return float(json['utilization'])

    return get_http_utilization_for_all_tasks(marathon_service_config, marathon_tasks, endpoint, utilization_mapper)


@register_autoscaling_component('mesos_cpu', SERVICE_METRICS_PROVIDER_KEY)
def mesos_cpu_metrics_provider(
    marathon_service_config, system_paasta_config, marathon_tasks, mesos_tasks, log_utilization_data={},
    noop=False, **kwargs,
):
    """
    Gets the mean cpu utilization of a service across all of its tasks.

    :param marathon_service_config: the MarathonServiceConfig to get data from
    :param marathon_tasks: Marathon tasks to get data from
    :param mesos_tasks: Mesos tasks to get data from
    :param log_utilization_data: A dict used to transfer utilization data to autoscale_marathon_instance()

    :returns: the service's mean utilization, from 0 to 1
    """

    autoscaling_root = compose_autoscaling_zookeeper_root(
        service=marathon_service_config.service,
        instance=marathon_service_config.instance,
    )
    zk_last_time_path = '%s/cpu_last_time' % autoscaling_root
    zk_last_cpu_data = '%s/cpu_data' % autoscaling_root

    with ZookeeperPool() as zk:
        try:
            last_time = zk.get(zk_last_time_path)[0].decode('utf8')
            last_cpu_data = zk.get(zk_last_cpu_data)[0].decode('utf8')
            log_utilization_data[last_time] = last_cpu_data
            last_time = float(last_time)
            last_cpu_data = (datum for datum in last_cpu_data.split(',') if datum)
        except NoNodeError:
            last_time = 0.0
            last_cpu_data = []

    futures = [asyncio.ensure_future(task.stats()) for task in mesos_tasks]
    if futures:
        a_sync.block(asyncio.wait, futures, timeout=60)

    def results_or_None(fut):
        if fut.exception():
            return None
        else:
            return fut.result()

    mesos_tasks_stats = dict(zip([task['id'] for task in mesos_tasks], [results_or_None(fut) for fut in futures]))

    current_time = int(datetime.now().strftime('%s'))
    time_delta = current_time - last_time

    mesos_cpu_data = {}
    for task_id, stats in mesos_tasks_stats.items():
        if stats is not None:
            try:
                utime = float(stats['cpus_user_time_secs'])
                stime = float(stats['cpus_system_time_secs'])
                limit = float(stats['cpus_limit']) - .1
                mesos_cpu_data[task_id] = (stime + utime) / limit
            except KeyError:
                pass

    if not mesos_cpu_data:
        raise MetricsProviderNoDataError("Couldn't get any cpu data from Mesos")

    cpu_data_csv = ','.join(f'{cpu_seconds}:{task_id}' for task_id, cpu_seconds in mesos_cpu_data.items())
    log_utilization_data[str(current_time)] = cpu_data_csv

    if not noop:
        with ZookeeperPool() as zk:
            zk.ensure_path(zk_last_cpu_data)
            zk.ensure_path(zk_last_time_path)
            zk.set(zk_last_cpu_data, str(cpu_data_csv).encode('utf8'))
            zk.set(zk_last_time_path, str(current_time).encode('utf8'))

    utilization = {}
    for datum in last_cpu_data:
        last_cpu_seconds, task_id = datum.split(':')
        if task_id in mesos_cpu_data:
            cputime_delta = mesos_cpu_data[task_id] - float(last_cpu_seconds)
            utilization[task_id] = cputime_delta / time_delta

    if not utilization:
        raise MetricsProviderNoDataError("""The mesos_cpu metrics provider doesn't have Zookeeper data for this service.
                                         This is expected for its first run.""")

    task_utilization = utilization.values()
    mean_utilization = mean(task_utilization)
    return mean_utilization


def get_error_from_utilization(utilization, setpoint, current_instances):
    """
    Consider scaling up if utilization is above the setpoint
    Consider scaling down if the utilization is below the setpoint AND scaling down wouldn't bring it above the setpoint
    Otherwise don't scale
    """
    max_threshold = setpoint
    min_threshold = max_threshold * (current_instances - 1) / current_instances
    if utilization < min_threshold:
        return utilization - min_threshold
    elif utilization > max_threshold:
        return utilization - max_threshold
    else:
        return 0.0


def get_autoscaling_info(apps_with_clients, service_config):
    if service_config.get_max_instances() and service_config.get_desired_state() == 'start':
        all_mesos_tasks = a_sync.block(get_cached_list_of_running_tasks_from_frameworks)
        autoscaling_params = service_config.get_autoscaling_params()
        autoscaling_params.update({'noop': True})
        system_paasta_config = load_system_paasta_config()
        try:
            marathon_tasks, mesos_tasks = filter_autoscaling_tasks(
                [app for (app, client) in apps_with_clients],
                all_mesos_tasks,
                service_config,
            )
            utilization = get_utilization(
                marathon_service_config=service_config,
                system_paasta_config=system_paasta_config,
                autoscaling_params=autoscaling_params,
                log_utilization_data={},
                marathon_tasks=list(marathon_tasks.values()),
                mesos_tasks=mesos_tasks,
            )
            error = get_error_from_utilization(
                utilization=utilization,
                setpoint=autoscaling_params['setpoint'],
                current_instances=service_config.get_instances(),
            )
            new_instance_count = get_new_instance_count(
                utilization=utilization,
                error=error,
                autoscaling_params=autoscaling_params,
                current_instances=service_config.get_instances(),
                marathon_service_config=service_config,
                num_healthy_instances=len(marathon_tasks),
            )
            current_utilization = "{:.1f}%".format(utilization * 100)
        except MetricsProviderNoDataError:
            current_utilization = "Exception"
            new_instance_count = "Exception"
        return ServiceAutoscalingInfo(
            current_instances=str(service_config.get_instances()),
            max_instances=str(service_config.get_max_instances()),
            min_instances=str(service_config.get_min_instances()),
            current_utilization=current_utilization,
            target_instances=str(new_instance_count),
        )
    return None


def get_new_instance_count(
    utilization, error, autoscaling_params, current_instances, marathon_service_config,
    num_healthy_instances,
):
    autoscaling_decision_policy = get_decision_policy(autoscaling_params[DECISION_POLICY_KEY])

    zookeeper_path = compose_autoscaling_zookeeper_root(
        service=marathon_service_config.service,
        instance=marathon_service_config.instance,
    )
    autoscaling_amount = autoscaling_decision_policy(
        utilization=utilization,
        error=error,
        min_instances=marathon_service_config.get_min_instances(),
        max_instances=marathon_service_config.get_max_instances(),
        current_instances=current_instances,
        zookeeper_path=zookeeper_path,
        num_healthy_instances=num_healthy_instances,
        **autoscaling_params,
    )

    # Limit downscaling by 30% of current_instances until we find out what is
    # going on in such situations
    safe_downscaling_threshold = int(current_instances * 0.7)
    new_instance_count = max(current_instances + autoscaling_amount, safe_downscaling_threshold)

    new_instance_count = marathon_service_config.limit_instance_count(new_instance_count)
    return new_instance_count


def get_utilization(
    marathon_service_config,
    system_paasta_config,
    autoscaling_params,
    log_utilization_data,
    marathon_tasks,
    mesos_tasks,
):
    autoscaling_metrics_provider = get_service_metrics_provider(autoscaling_params[SERVICE_METRICS_PROVIDER_KEY])

    return autoscaling_metrics_provider(
        marathon_service_config=marathon_service_config,
        system_paasta_config=system_paasta_config,
        marathon_tasks=marathon_tasks,
        mesos_tasks=mesos_tasks,
        log_utilization_data=log_utilization_data,
        **autoscaling_params,
    )


def is_task_data_insufficient(marathon_service_config, marathon_tasks, current_instances):
    too_many_instances_running = len(marathon_tasks) > int((1 + MAX_TASK_DELTA) * current_instances)
    too_few_instances_running = len(marathon_tasks) < int((1 - MAX_TASK_DELTA) * current_instances)
    return too_many_instances_running or too_few_instances_running


def autoscale_marathon_instance(
    marathon_service_config: MarathonServiceConfig,
    system_paasta_config: SystemPaastaConfig,
    marathon_tasks: Sequence[MarathonTask],
    mesos_tasks: Sequence[Task],
) -> None:
    try:
        with create_autoscaling_lock(marathon_service_config.service, marathon_service_config.instance):
            current_instances = marathon_service_config.get_instances()
            task_data_insufficient = is_task_data_insufficient(
                marathon_service_config=marathon_service_config,
                marathon_tasks=marathon_tasks,
                current_instances=current_instances,
            )
            autoscaling_params = marathon_service_config.get_autoscaling_params()
            log_utilization_data: Mapping = {}
            utilization = get_utilization(
                marathon_service_config=marathon_service_config,
                system_paasta_config=system_paasta_config,
                autoscaling_params=autoscaling_params,
                log_utilization_data=log_utilization_data,
                marathon_tasks=marathon_tasks,
                mesos_tasks=mesos_tasks,
            )
            error = get_error_from_utilization(
                utilization=utilization,
                setpoint=autoscaling_params['setpoint'],
                current_instances=current_instances,
            )
            num_healthy_instances = len(marathon_tasks)
            new_instance_count = get_new_instance_count(
                utilization=utilization,
                error=error,
                autoscaling_params=autoscaling_params,
                current_instances=current_instances,
                marathon_service_config=marathon_service_config,
                num_healthy_instances=num_healthy_instances,
            )
            safe_downscaling_threshold = int(current_instances * 0.7)
            _record_autoscaling_decision(
                marathon_service_config=marathon_service_config,
                autoscaling_params=autoscaling_params,
                utilization=utilization,
                log_utilization_data=log_utilization_data,
                error=error,
                current_instances=current_instances,
                num_healthy_instances=num_healthy_instances,
                new_instance_count=new_instance_count,
                safe_downscaling_threshold=safe_downscaling_threshold,
                task_data_insufficient=task_data_insufficient,
            )
            if new_instance_count != current_instances:
                if new_instance_count < current_instances and task_data_insufficient:
                    write_to_log(
                        config=marathon_service_config,
                        line='Delaying scaling *down* as we found too few healthy tasks running in marathon. '
                        'This can happen because tasks are delayed/waiting/unhealthy or because we are '
                        'waiting for tasks to be killed. Will wait for sufficient healthy tasks before '
                        'we make a decision to scale down.',
                        level='debug',
                    )
                    return
                else:
                    set_instances_for_marathon_service(
                        service=marathon_service_config.service,
                        instance=marathon_service_config.instance,
                        instance_count=new_instance_count,
                    )
                    write_to_log(
                        config=marathon_service_config,
                        line='Scaling from %d to %d instances (%s)' % (
                            current_instances, new_instance_count, humanize_error(error),
                        ),
                        level='event',
                    )
    except LockHeldException:
        log.warning("Skipping autoscaling run for {service}.{instance} because the lock is held".format(
            service=marathon_service_config.service,
            instance=marathon_service_config.instance,
        ))


def _record_autoscaling_decision(
    marathon_service_config: MarathonServiceConfig,
    autoscaling_params: AutoscalingParamsDict,
    utilization: float,
    log_utilization_data: Mapping[str, str],
    error: float,
    current_instances: int,
    num_healthy_instances: int,
    new_instance_count: int,
    safe_downscaling_threshold: int,
    task_data_insufficient: bool,
) -> None:
    """
    Based on the calculations made, perform observability side effects.
    Log messages, generate time series, send any alerts, etc.
    """
    write_to_log(
        config=marathon_service_config,
        line=json.dumps(
            dict(
                timestamp=time.time(),
                paasta_cluster=marathon_service_config.get_cluster(),
                paasta_service=marathon_service_config.get_service(),
                paasta_instance=marathon_service_config.get_instance(),
                autoscaling_params=autoscaling_params,
                utilization=utilization,
                error=error,
                current_instances=current_instances,
                num_healthy_instances=num_healthy_instances,
                new_instance_count=new_instance_count,
                safe_downscaling_threshold=safe_downscaling_threshold,
                task_data_insufficient=task_data_insufficient,
            ),
        ),
        level='debug',
    )
    meteorite_dims = {
        'service_name': marathon_service_config.service,
        'decision_policy': autoscaling_params[DECISION_POLICY_KEY],  # type: ignore
        'paasta_cluster': marathon_service_config.cluster,
        'instance_name': marathon_service_config.instance,
    }
    if yelp_meteorite:
        gauge = yelp_meteorite.create_gauge('paasta.service.instances', meteorite_dims)
        gauge.set(new_instance_count)
        gauge = yelp_meteorite.create_gauge('paasta.service.max_instances', meteorite_dims)
        gauge.set(marathon_service_config.get_max_instances())
        gauge = yelp_meteorite.create_gauge('paasta.service.min_instances', meteorite_dims)
        gauge.set(marathon_service_config.get_min_instances())


def humanize_error(error):
    if error < 0:
        return '%d%% underutilized' % floor(-error * 100)
    elif error > 0:
        return '%d%% overutilized' % ceil(error * 100)
    else:
        return 'utilization within thresholds'


def get_configs_of_services_to_scale(
    cluster: str,
    soa_dir: str = DEFAULT_SOA_DIR,
    services: Optional[Sequence[str]] = None,
) -> Sequence[MarathonServiceConfig]:
    if not services:
        services = list_services(
            soa_dir=soa_dir,
        )
    configs = []
    for service in services:
        service_config = PaastaServiceConfigLoader(
            service=service,
            soa_dir=soa_dir,
        )
        for instance_config in service_config.instance_configs(
            cluster=cluster,
            instance_type_class=MarathonServiceConfig,
        ):
            if instance_config.get_max_instances() and instance_config.get_desired_state() == 'start' \
                    and instance_config.get_autoscaling_params()['decision_policy'] != 'bespoke':
                configs.append(instance_config)

    return configs


def autoscaling_is_paused():
    with ZookeeperPool() as zk:
        try:
            pause_until = zk.get(ZK_PAUSE_AUTOSCALE_PATH)[0].decode('utf8')
            pause_until = float(pause_until)
        except (NoNodeError, ValueError, AttributeError):
            pause_until = 0

    remaining = pause_until - time.time()
    if remaining >= 0:
        log.debug("Autoscaling is paused for {} more seconds".format(str(remaining)))
        return True
    else:
        return False


def autoscale_services(
    soa_dir: str = DEFAULT_SOA_DIR,
    services: Optional[Sequence[str]] = None,
) -> None:
    system_paasta_config = load_system_paasta_config()
    cluster = system_paasta_config.get_cluster()
    configs = get_configs_of_services_to_scale(cluster=cluster, soa_dir=soa_dir, services=services)
    autoscale_service_configs(service_configs=configs, system_paasta_config=system_paasta_config)


@use_requests_cache('service_autoscaler')
def autoscale_service_configs(
    service_configs: Sequence[MarathonServiceConfig],
    system_paasta_config: SystemPaastaConfig,
) -> None:
    if autoscaling_is_paused():
        log.warning("Skipping autoscaling because autoscaler paused")
        return

    marathon_clients = get_marathon_clients(get_marathon_servers(system_paasta_config))
    apps_with_clients = get_marathon_apps_with_clients(marathon_clients.get_all_clients(), embed_tasks=True)
    all_mesos_tasks = a_sync.block(get_all_running_tasks)
    with ZookeeperPool():
        for config in service_configs:
            try:
                marathon_tasks, mesos_tasks = filter_autoscaling_tasks(
                    [app for (app, client) in apps_with_clients],
                    all_mesos_tasks,
                    config,
                )
                autoscale_marathon_instance(
                    config,
                    system_paasta_config,
                    list(marathon_tasks.values()),
                    mesos_tasks,
                )
            except Exception as e:
                write_to_log(
                    config=config,
                    line='Caught Exception %s' % e,
                    level='debug',
                )


def filter_autoscaling_tasks(
    marathon_apps: Sequence[MarathonApp],
    all_mesos_tasks: Sequence[Task],
    config: MarathonServiceConfig,
) -> Tuple[Mapping[str, MarathonTask], Sequence[Task]]:
    job_id_prefix = "{}{}".format(format_job_id(service=config.service, instance=config.instance), MESOS_TASK_SPACER)

    # Get a dict of healthy tasks, we assume tasks with no healthcheck defined
    # are healthy. We assume tasks with no healthcheck results but a defined
    # healthcheck to be unhealthy (unless they are "old" in which case we
    # assume that marathon has screwed up and stopped healthchecking but that
    # they are healthy
    log.info("Inspecting %s for autoscaling" % job_id_prefix)
    marathon_tasks = {}
    for app in marathon_apps:
        for task in app.tasks:
            if task.id.startswith(job_id_prefix) and (
                is_task_healthy(task) or not
                app.health_checks or is_old_task_missing_healthchecks(task, app)
            ):
                marathon_tasks[task.id] = task

    if not marathon_tasks:
        raise MetricsProviderNoDataError("Couldn't find any healthy marathon tasks")
    mesos_tasks = [task for task in all_mesos_tasks if task['id'] in marathon_tasks]
    return (marathon_tasks, mesos_tasks)


def write_to_log(config, line, level='event'):
    _log(
        service=config.service,
        line=line,
        component='deploy',
        level=level,
        cluster=config.cluster,
        instance=config.instance,
    )


def get_short_job_id(task_id):
    return MESOS_TASK_SPACER.join(task_id.split(MESOS_TASK_SPACER, 2)[:2])


@contextmanager
def create_autoscaling_lock(service: str, instance: str) -> Iterator[None]:
    """Acquire a lock in zookeeper for autoscaling. This is
    to avoid autoscaling a service multiple times, and to avoid
    having multiple paasta services all attempting to autoscale and
    fetching mesos data."""
    zk = KazooClient(hosts=load_system_paasta_config().get_zk_hosts(), timeout=ZK_LOCK_CONNECT_TIMEOUT_S)
    zk.start()
    lock = zk.Lock(f'/autoscaling/{service}/{instance}/autoscaling.lock')
    try:
        lock.acquire(timeout=1)  # timeout=0 throws some other strange exception
        yield
    except LockTimeout:
        raise LockHeldException(f"Failed to acquire lock for autoscaling! {service}.{instance}")
    else:
        lock.release()
    finally:
        zk.stop()
        zk.close()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import shlex
import subprocess
import sys
from urllib.parse import urlparse

from paasta_tools.cli.utils import calculate_remote_masters
from paasta_tools.cli.utils import find_connectable_master
from paasta_tools.cli.utils import get_status_for_instance
from paasta_tools.cli.utils import get_subparser
from paasta_tools.cli.utils import pick_slave_from_status
from paasta_tools.marathon_tools import get_marathon_clients
from paasta_tools.marathon_tools import get_marathon_servers
from paasta_tools.marathon_tools import load_marathon_service_config
from paasta_tools.mesos_tools import get_mesos_master
from paasta_tools.utils import _run
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print


def add_subparser(subparsers):
    new_parser = get_subparser(
        description="'paasta sysdig' works by SSH'ing to remote PaaSTA masters and "
                    "running sysdig with the necessary filters",
        help_text="Run sysdig on a remote host and filter to a service and instance",
        command='sysdig',
        function=paasta_sysdig,
        subparsers=subparsers,
    )
    new_parser.add_argument(
        '-l', '--local',
        help="Run the script here rather than SSHing to a PaaSTA master",
        default=False,
        action='store_true',
    )


def get_any_mesos_master(cluster, system_paasta_config):
    masters, output = calculate_remote_masters(cluster, system_paasta_config)
    if not masters:
        paasta_print('ERROR: %s' % output)
        sys.exit(1)
    mesos_master, output = find_connectable_master(masters)
    if not mesos_master:
        paasta_print(f'ERROR: could not find connectable master in cluster {cluster}\nOutput: {output}')
        sys.exit(1)
    return mesos_master


def paasta_sysdig(args):
    system_paasta_config = load_system_paasta_config()

    if not args.local:
        mesos_master = get_any_mesos_master(cluster=args.cluster, system_paasta_config=system_paasta_config)
        ssh_cmd = (
            'ssh -At -o StrictHostKeyChecking=no -o LogLevel=QUIET {0} '
            '"sudo paasta {1} --local"'
        ).format(mesos_master, ' '.join(sys.argv[1:]))
        return_code, output = _run(ssh_cmd)
        if return_code != 0:
            paasta_print(output)
            sys.exit(return_code)
        slave, command = output.split(':', 1)
        subprocess.call(shlex.split("ssh -tA {} '{}'".format(slave, command.strip())))
        return
    status = get_status_for_instance(
        cluster=args.cluster,
        service=args.service,
        instance=args.instance,
    )
    slave = pick_slave_from_status(
        status=status,
        host=args.host,
    )

    job_config = load_marathon_service_config(
        service=args.service,
        instance=args.instance,
        cluster=args.cluster,
    )

    marathon_servers = get_marathon_servers(system_paasta_config)
    marathon_clients = get_marathon_clients(marathon_servers)

    # Unfortunately, sysdig seems to only be able to take one marathon URL, so hopefully the service in question is not
    # currently moving between shards.
    client = marathon_clients.get_current_client_for_service(
        job_config=job_config,
    )
    marathon_url = client.servers[0]
    marathon_user, marathon_pass = client.auth

    mesos_url = get_mesos_master().host
    marathon_parsed_url = urlparse(marathon_url)
    marathon_creds_url = marathon_parsed_url._replace(netloc="{}:{}@{}".format(
        marathon_user, marathon_pass,
        marathon_parsed_url.netloc,
    ))
    paasta_print(format_mesos_command(slave, status.marathon.app_id, mesos_url, marathon_creds_url.geturl()))


def format_mesos_command(slave, app_id, mesos_url, marathon_url):
    sysdig_mesos = f'{mesos_url},{marathon_url}'
    command = f'sudo csysdig -m {sysdig_mesos} marathon.app.id="/{app_id}" -v mesos_tasks'
    return slave + ":" + command



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Usage: ./list_marathon_service_instances.py [options]

Enumerates all marathon instances for services in the soa directory that
are for the current cluster (defined by the marathon configuration file).

Outputs (to stdout) a list of service.instance (one per line)
for each instance found in marathon-<CLUSTER>.yaml for every folder
in the SOA Configuration directory.

Command line options:

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
- -c <CLUSTER>, --cluster <CLUSTER>: Specify which cluster of services to read
- -m, --minimal: Only show service instances that need bouncing
"""
import argparse
import sys

from paasta_tools.marathon_tools import DEFAULT_SOA_DIR
from paasta_tools.marathon_tools import get_marathon_apps_with_clients
from paasta_tools.marathon_tools import get_marathon_clients
from paasta_tools.marathon_tools import get_marathon_servers
from paasta_tools.marathon_tools import get_num_at_risk_tasks
from paasta_tools.marathon_tools import load_marathon_service_config
from paasta_tools.mesos_maintenance import get_draining_hosts
from paasta_tools.utils import _log
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import get_services_for_cluster
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import long_job_id_to_short_job_id
from paasta_tools.utils import paasta_print
from paasta_tools.utils import use_requests_cache


def parse_args():
    parser = argparse.ArgumentParser(
        description='Lists marathon instances for a service.',
    )
    parser.add_argument(
        '-c', '--cluster', dest="cluster", metavar="CLUSTER",
        default=None,
        help="define a specific cluster to read from",
    )
    parser.add_argument(
        '-d', '--soa-dir', dest="soa_dir", metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        '-m', '--minimal', dest='minimal', action='store_true',
        help="show only service instances that need bouncing",
    )
    args = parser.parse_args()
    return args


def get_desired_marathon_configs(soa_dir):
    cluster = load_system_paasta_config().get_cluster()
    instances = get_services_for_cluster(
        instance_type='marathon',
        cluster=cluster,
        soa_dir=soa_dir,
    )

    job_configs = dict()
    formatted_marathon_configs = dict()

    for service, instance in instances:
        try:
            job_config = load_marathon_service_config(
                service=service,
                instance=instance,
                cluster=cluster,
                soa_dir=soa_dir,
            )

            formatted_config = job_config.format_marathon_app_dict()
            formatted_marathon_configs[formatted_config['id'].lstrip('/')] = formatted_config
            job_configs[formatted_config['id'].lstrip('/')] = job_config
        # Not ideal but we rely on a lot of user input to create the app dict
        # and we really can't afford to bail if just one app definition is malformed
        except Exception as errormsg:
            _log(
                service=service,
                line=str(errormsg),
                component='deploy',
                level='debug',
                cluster=cluster,
                instance=instance,
            )
    return formatted_marathon_configs, job_configs


@use_requests_cache('list_marathon_services')
def get_service_instances_that_need_bouncing(marathon_clients, soa_dir):
    desired_marathon_configs_formatted, desired_job_configs = get_desired_marathon_configs(soa_dir)
    desired_ids_and_clients = set()
    for app_id, job_config in desired_job_configs.items():
        desired_ids_and_clients.add((app_id, marathon_clients.get_current_client_for_service(job_config)))

    current_apps_with_clients = {
        (app.id.lstrip('/'), client): app
        for app, client in get_marathon_apps_with_clients(marathon_clients.get_all_clients())
    }
    actual_ids_and_clients = set(current_apps_with_clients.keys())

    undesired_apps_and_clients = actual_ids_and_clients.symmetric_difference(desired_ids_and_clients)
    apps_that_need_bouncing = {long_job_id_to_short_job_id(app_id) for app_id, client in undesired_apps_and_clients}

    draining_hosts = get_draining_hosts()

    for (app_id, client), app in current_apps_with_clients.items():
        short_app_id = long_job_id_to_short_job_id(app_id)
        if short_app_id not in apps_that_need_bouncing:
            if (
                app.instances != desired_marathon_configs_formatted[app_id]['instances'] or
                get_num_at_risk_tasks(app, draining_hosts) != 0
            ):
                apps_that_need_bouncing.add(short_app_id)

    return (app_id.replace('--', '_') for app_id in apps_that_need_bouncing)


def main():
    args = parse_args()
    soa_dir = args.soa_dir
    cluster = args.cluster
    if args.minimal:
        system_paasta_config = load_system_paasta_config()
        marathon_servers = get_marathon_servers(system_paasta_config)
        marathon_clients = get_marathon_clients(marathon_servers)
        service_instances = get_service_instances_that_need_bouncing(
            marathon_clients=marathon_clients, soa_dir=soa_dir,
        )
    else:
        instances = get_services_for_cluster(
            cluster=cluster,
            instance_type='marathon',
            soa_dir=soa_dir,
        )
        service_instances = []
        for name, instance in instances:
            service_instances.append(compose_job_id(name, instance))
    paasta_print('\n'.join(service_instances))
    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import argparse
import datetime
import logging
from collections import defaultdict

import pysensu_yelp
import requests

from paasta_tools import mesos_tools
from paasta_tools.monitoring_tools import send_event
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import list_services


logger = logging.getLogger(__name__)


JUPYTER_PREFIX = 'jupyterhub_'


def parse_args():
    parser = argparse.ArgumentParser(
        description='Reports long-running Spark frameworks.',
    )
    parser.add_argument(
        '--min-hours',
        type=float,
        help='Report frameworks that have been registered for more than this duration',
        default=0,
    )
    parser.add_argument(
        '--no-notify',
        action='store_true',
        help='Skip notifying the teams that own each framework',
    )
    return parser.parse_args()


def get_time_running(framework):
    registered_time = datetime.datetime.fromtimestamp(framework['registered_time'])
    return datetime.datetime.now() - registered_time


def get_spark_properties(framework):
    webui_url = framework.get('webui_url')
    if not webui_url:
        return None

    env_endpoint = f'{webui_url}/api/v1/applications/{framework.id}/environment'
    try:
        response = requests.get(env_endpoint, timeout=5)
    except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as e:
        logger.warning(f'Unable to connect to {env_endpoint}: {e!r}')
        return None

    if response.status_code != 200:
        logger.warning(f'Bad response from {env_endpoint}: {response.status_code}')
        return None

    try:
        return response.json()['sparkProperties']
    except (ValueError, KeyError):
        logger.warning(f'Unable to get sparkProperties for {framework.id}: got response {response.text}')
        return None


def guess_service(properties):
    if not properties:
        return None
    for key, value in properties:
        if key == 'spark.executorEnv.PAASTA_SERVICE':
            service = value
            break
    else:
        return None
    if service.startswith(JUPYTER_PREFIX):
        return service[len(JUPYTER_PREFIX):]
    else:
        return service


def get_matching_framework_info(min_hours):
    frameworks = mesos_tools.get_all_frameworks(active_only=True)
    matching_info = []
    min_timedelta = datetime.timedelta(hours=min_hours)
    for framework in frameworks:
        if not framework.active:
            continue
        if framework.get('principal') != 'spark':
            continue
        time_running = get_time_running(framework)
        if time_running >= min_timedelta:
            info = {
                'id': framework.id,
                'name': framework.name,
                'webui_url': framework.get('webui_url'),
                'service': guess_service(get_spark_properties(framework)),
                'user': framework.user,
                'time_running': str(time_running),
            }
            matching_info.append(info)

    return matching_info


def format_framework(info):
    result = [f'{info["name"]} (running for {info["time_running"]})']
    result.append(f'  user: {info["user"]}')
    result.append(f'  job UI: {info["webui_url"]}')
    return '\n'.join(result)


def format_message_for_service(service, frameworks):
    output = f'Found the following long-running Spark frameworks associated with service {service}.\n'
    output += f'Please check why they are still running and terminate if appropriate.\n\n'
    output += '\n'.join(format_framework(f) for f in frameworks)
    return output


def get_messages_by_service(results):
    results_by_service = defaultdict(list)
    for result in results:
        service = result['service']
        results_by_service[service].append(result)

    return {
        service: format_message_for_service(service, results)
        for service, results in results_by_service.items()
    }


def update_check_status(service, output, status):
    overrides = {
        'page': False,
        'alert_after': 0,
        'tip': 'Ask the user to check the job UI and terminate the job if appropriate.',
        'runbook': 'http://y/spark-debug',
        'ticket': True,
    }
    send_event(
        service=service,
        check_name=f'long_running_spark_jobs.{service}',
        overrides=overrides,
        status=status,
        output=output,
        soa_dir=DEFAULT_SOA_DIR,
    )


def report_spark_jobs(min_hours, no_notify):
    results = get_matching_framework_info(
        min_hours=min_hours,
    )
    messages_by_service = get_messages_by_service(results)
    valid_services = set(list_services())

    messages_for_unknown_services = []
    for service, message in messages_by_service.items():
        if service in valid_services:
            print(f'{message}\n')
        else:
            messages_for_unknown_services.append(message)
    if messages_for_unknown_services:
        print('\nINVALID SERVICES')
        print('----------------')
        print('The following frameworks are associated with services that are not configured in PaaSTA.\n')
        print('\n\n'.join(messages_for_unknown_services))

    if not no_notify:
        for service in valid_services:
            if service in messages_by_service:
                update_check_status(service, message, pysensu_yelp.Status.WARNING)
            else:
                update_check_status(service, 'No long running spark jobs', pysensu_yelp.Status.OK)


def main():
    args = parse_args()
    logging.basicConfig()
    report_spark_jobs(args.min_hours, args.no_notify)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import errno
import json
import os.path
import random
import shlex
import sys
import threading
from subprocess import PIPE
from subprocess import Popen
from subprocess import STDOUT

import yaml

from paasta_tools.utils import paasta_print


def _timeout(process):
    """Helper function for _run. It terminates the process.
    Doesn't raise OSError, if we try to terminate a non-existing
    process as there can be a very small window between poll() and kill()
    """
    if process.poll() is None:
        try:
            # sending SIGKILL to the process
            process.kill()
        except OSError as e:
            # No such process error
            # The process could have been terminated meanwhile
            if e.errno != errno.ESRCH:
                raise


def cmd(command):
    stream = False
    timeout = 60
    output = []
    try:
        process = Popen(shlex.split(command), stdout=PIPE, stderr=STDOUT, stdin=None)
        process.name = command
        # start the timer if we specified a timeout
        if timeout:
            proctimer = threading.Timer(timeout, _timeout, (process,))
            proctimer.start()
        for line in iter(process.stdout.readline, ''):
            if stream:
                paasta_print(line.rstrip('\n'))
            output.append(line.rstrip('\n'))
        # when finished, get the exit code
        returncode = process.wait()
    except OSError as e:
        output.append(e.strerror.rstrip('\n'))
        returncode = e.errno
    except (KeyboardInterrupt, SystemExit):
        # need to clean up the timing thread here
        if timeout:
            proctimer.cancel()
        raise
    else:
        # Stop the timer
        if timeout:
            proctimer.cancel()
    if returncode == -9:
        output.append(f"Command '{command}' timed out (longer than {timeout}s)")
    return returncode, '\n'.join(output)


def abort(message):
    paasta_print(message)
    sys.exit(1)


def condquit(rc, message):
    if rc != 0:
        paasta_print(message)
        sys.exit(rc)


def docker_env_to_dict(environment_array):
    environment = {}
    for kv in environment_array:
        k, v = kv.split('=', 1)
        environment[k] = v
    return environment


def get_proxy_port(service_name, instance_name):
    smartstack_yaml = "/nail/etc/services/%s/smartstack.yaml" % service_name
    proxy_port = None
    if os.path.exists(smartstack_yaml):
        with open(smartstack_yaml, 'r') as stream:
            data = yaml.safe_load(stream)
            if instance_name in data:
                proxy_port = data[instance_name].get('proxy_port', None)
    return proxy_port


def get_last_killed(drained_apps, service, instance):
    """look "back" in drained_apps, find at what time
    the given (service, instance) was last killed"""
    last_killed_t = -1000
    for drained_app in reversed(drained_apps):
        dt, dservice, dinstance = drained_app
        if dservice == service and dinstance == instance:
            last_killed_t = dt
            break
    return last_killed_t


def has_all_paasta_env(environment):
    for k in ('PAASTA_SERVICE', 'PAASTA_INSTANCE', 'MARATHON_PORT'):
        if k not in environment:
            return False

    return True


def main():
    rc, output = cmd('sudo docker ps -q')
    condquit(rc, 'docker ps')
    lines = output.split("\n")

    if len(lines) == 0:
        abort("no containers running")

    running_container_ids = []

    for line in lines:
        if len(line) != 12:
            abort("%s doesn't look like a container ID" % line)
        running_container_ids.append(line.rstrip())

    random.shuffle(running_container_ids)

    drained_apps = []  # ( t_killed, service, instance )
    smartstack_grace_sleep = 10
    between_containers_grace_sleep = 10
    min_kill_interval = 60  # minimum time to wait between same service.instance kills
    hadown_expire_in_seconds = 120
    t = 0

    for container_id in running_container_ids:
        rc, output = cmd("sudo docker inspect %s" % container_id)
        condquit(rc, "docker inspect %s" % container_id)
        docker_inspect_data = json.loads(output)
        environment = docker_env_to_dict(docker_inspect_data[0]['Config']['Env'])
        if 'CHRONOS_JOB_NAME' in environment:
            paasta_print(
                "# WARNING! {} is a chronos job ({}), skipping".format(container_id, environment['CHRONOS_JOB_NAME']),
            )
            paasta_print()
            continue
        if not has_all_paasta_env(environment):
            paasta_print("# WARNING: %s is not a paasta container, skipping)" % (container_id))
            continue
        service = environment['PAASTA_SERVICE']
        instance = environment['PAASTA_INSTANCE']
        paasta_print(f"# {service}.{instance}")
        marathon_port = int(environment['MARATHON_PORT'])
        proxy_port = get_proxy_port(service, instance)
        paasta_print(f"# {container_id},{service},{instance},{proxy_port},{marathon_port}")
        paasta_print("sudo hadown -P {} -e $((`date +'%s'`+{})) {}.{}".format(
            marathon_port, hadown_expire_in_seconds, service, instance,
        ))
        paasta_print("sleep %s" % smartstack_grace_sleep)
        t += smartstack_grace_sleep
        paasta_print("sudo docker kill %s" % container_id)
        paasta_print(f"sudo haup -P {marathon_port} {service}.{instance}")
        last_killed_t = get_last_killed(drained_apps, service, instance)
        drained_apps.append((t, service, instance))
        # print "t:%s last_killed_t:%s" % (t, last_killed_t)
        sleep_amount = between_containers_grace_sleep
        if (t - last_killed_t) < min_kill_interval:
            sleep_amount = min_kill_interval - (t - last_killed_t) + between_containers_grace_sleep
        paasta_print("sleep %s" % sleep_amount)
        t += sleep_amount
        paasta_print()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains methods used by the paasta client to mark a docker image for
deployment to a cluster.instance.
"""
import getpass
import logging
import math
import os
import socket
import sys
import time
import traceback
from collections import defaultdict
from queue import Empty
from queue import Queue
from threading import Event
from threading import Thread
from typing import Collection
from typing import Iterator
from typing import Mapping
from typing import Optional

import progressbar
from bravado.exception import HTTPError
from requests.exceptions import ConnectionError
from service_configuration_lib import read_deploy
from slackclient import SlackClient

from paasta_tools import remote_git
from paasta_tools.api import client
from paasta_tools.automatic_rollbacks import state_machine
from paasta_tools.automatic_rollbacks.slo import SLOSlackDeploymentProcess
from paasta_tools.cli.cmds.push_to_registry import is_docker_image_already_in_registry
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_deploy_groups
from paasta_tools.cli.utils import validate_git_sha
from paasta_tools.cli.utils import validate_given_deploy_groups
from paasta_tools.cli.utils import validate_service_name
from paasta_tools.cli.utils import validate_short_git_sha
from paasta_tools.deployment_utils import get_currently_deployed_sha
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.marathon_tools import MarathonServiceConfig
from paasta_tools.paasta_service_config_loader import PaastaServiceConfigLoader
from paasta_tools.slack import get_slack_client
from paasta_tools.utils import _log
from paasta_tools.utils import _log_audit
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import format_tag
from paasta_tools.utils import get_git_url
from paasta_tools.utils import get_paasta_tag_from_deploy_group
from paasta_tools.utils import list_services
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import TimeoutError


DEFAULT_DEPLOYMENT_TIMEOUT = 3600  # seconds
DEFAULT_AUTO_CERTIFY_DELAY = 600  # seconds
DEFAULT_SLACK_CHANNEL = "#deploy"

log = logging.getLogger(__name__)


def add_subparser(subparsers):
    list_parser = subparsers.add_parser(
        'mark-for-deployment',
        help='Mark a docker image for deployment in git',
        description=(
            "'paasta mark-for-deployment' uses Git as the control-plane, to "
            "signal to other PaaSTA components that a particular docker image "
            "is ready to be deployed."
        ),
        epilog=(
            "Note: Access and credentials to the Git repo of a service are required "
            "for this command to work."
        ),
    )
    list_parser.add_argument(
        '-u', '--git-url',
        help=(
            'Git url for service -- where magic mark-for-deployment tags are pushed. '
            'Defaults to the normal git URL for the service.'
        ),
        default=None,
    )
    list_parser.add_argument(
        '-c', '-k', '--commit',
        help='Git sha to mark for deployment',
        required=True,
        type=validate_short_git_sha,
    )
    list_parser.add_argument(
        '-l', '--deploy-group', '--clusterinstance',
        help='Mark the service ready for deployment in this deploy group (e.g. '
             'cluster1.canary, cluster2.main). --clusterinstance is deprecated and '
             'should be replaced with --deploy-group',
        required=True,
    ).completer = lazy_choices_completer(list_deploy_groups)
    list_parser.add_argument(
        '-s', '--service',
        help='Name of the service which you wish to mark for deployment. Leading '
        '"services-" will be stripped.',
        required=True,
    ).completer = lazy_choices_completer(list_services)
    list_parser.add_argument(
        '--verify-image-exists',
        help='Check the docker registry and verify the image has been pushed',
        dest='verify_image',
        action='store_true',
        default=False,
    )
    list_parser.add_argument(
        '--wait-for-deployment',
        help='Set to poll paasta and wait for the deployment to finish, '
             'the default strategy is to mark for deployment and exit straightaway',
        dest='block',
        action='store_true',
        default=False,
    )
    list_parser.add_argument(
        '-t', '--timeout',
        dest="timeout",
        type=int,
        default=DEFAULT_DEPLOYMENT_TIMEOUT,
        help=(
            "Time in seconds to wait for paasta to deploy the service. "
            "If the timeout is exceeded we return 1. "
            "Default is %(default)s seconds."
        ),
    )
    list_parser.add_argument(
        '--auto-rollback',
        help='Automatically roll back to the previously deployed sha if the deployment '
             'times out or is canceled (ctrl-c). Only applicable with --wait-for-deployment. '
             'Defaults to false.',
        dest='auto_rollback',
        action='store_true',
        default=False,
    )
    list_parser.add_argument(
        '-d', '--soa-dir',
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    list_parser.add_argument(
        '-v', '--verbose',
        action='count',
        dest="verbose",
        default=0,
        help="Print out more output.",
    )
    list_parser.add_argument(
        '--auto-certify-delay',
        dest='auto_certify_delay',
        type=int,
        default=None,  # the logic for this is complicated. See MarkForDeploymentProcess.get_auto_certify_delay.
        help="After a deploy finishes, wait this many seconds before automatically certifying."
             f"Default {DEFAULT_AUTO_CERTIFY_DELAY} when --auto-rollback is enabled",
    )
    list_parser.add_argument(
        '--auto-abandon-delay',
        dest='auto_abandon_delay',
        type=int,
        default=600,
        help="After a rollback finishes, wait this many seconds before automatically abandoning.",
    )
    list_parser.add_argument(
        '--auto-rollback-delay',
        dest='auto_rollback_delay',
        type=int,
        default=30,
        help="After noticing an SLO failure, wait this many seconds before automatically rolling back.",
    )

    list_parser.set_defaults(command=paasta_mark_for_deployment)


def mark_for_deployment(git_url, deploy_group, service, commit):
    """Mark a docker image for deployment"""
    tag = get_paasta_tag_from_deploy_group(identifier=deploy_group, desired_state='deploy')
    remote_tag = format_tag(tag)
    ref_mutator = remote_git.make_force_push_mutate_refs_func(
        targets=[remote_tag],
        sha=commit,
    )

    max_attempts = 3
    for attempt in range(1, max_attempts + 1):
        try:
            remote_git.create_remote_refs(git_url=git_url, ref_mutator=ref_mutator, force=True)
        except Exception:
            logline = "Failed to mark {} for deployment in deploy group {}! (attempt {}/{})".format(
                commit, deploy_group, attempt, max_attempts,
            )
            _log(
                service=service,
                line=logline,
                component='deploy',
                level='event',
            )
            time.sleep(5 * attempt)
        else:
            logline = f"Marked {commit} for deployment in deploy group {deploy_group}"
            _log(
                service=service,
                line=logline,
                component='deploy',
                level='event',
            )

            audit_action_details = {
                'deploy_group': deploy_group,
                'commit': commit,
            }
            _log_audit(
                action='mark-for-deployment',
                action_details=audit_action_details,
                service=service,
            )

            return 0
    return 1


def report_waiting_aborted(service, deploy_group):
    paasta_print(PaastaColors.red(
        "Waiting for deployment aborted."
        " PaaSTA will continue trying to deploy this code.",
    ))
    paasta_print("If you wish to see the status, run:")
    paasta_print()
    paasta_print(f"    paasta status -s {service} -l {deploy_group} -v")
    paasta_print()


def get_authors_to_be_notified(git_url, from_sha, to_sha):
    if from_sha is None:
        return ""
    ret, authors = remote_git.get_authors(
        git_url=git_url, from_sha=from_sha, to_sha=to_sha,
    )
    if ret == 0:
        if authors == "":
            return ""
        else:
            slacky_authors = ", ".join([f"<@{a}>" for a in authors.split()])
            log.debug(f"Authors: {slacky_authors}")
            return f"Authors: {slacky_authors}"
    else:
        return f"(Could not get authors: {authors})"


class SlackDeployNotifier:
    def __init__(self, service, deploy_info, deploy_group, commit, old_commit, git_url, auto_rollback=False):
        self.sc = get_slack_client()
        self.service = service
        self.deploy_info = deploy_info
        self.deploy_group = deploy_group
        self.channels = deploy_info.get('slack_channels', [])
        self.commit = commit
        self.old_commit = self.lookup_production_deploy_group_sha() or old_commit
        self.git_url = git_url
        self.authors = get_authors_to_be_notified(git_url=self.git_url, from_sha=self.old_commit, to_sha=self.commit)
        self.url_message = self.get_url_message()
        self.auto_rollback = auto_rollback

    def get_url_message(self):
        build_url = os.environ.get('BUILD_URL')
        if build_url is not None:
            message = f"<{build_url}/consoleFull|Jenkins Job>"
        else:
            message = f"(Run by {getpass.getuser()} on {socket.getfqdn()})"
        return message

    def lookup_production_deploy_group_sha(self):
        prod_deploy_group = self.deploy_info.get('production_deploy_group', None)
        if prod_deploy_group is None:
            return None
        else:
            return get_currently_deployed_sha(service=self.service, deploy_group=prod_deploy_group)

    def deploy_group_is_set_to_notify(self, notify_type):
        for step in self.deploy_info.get('pipeline', []):
            if step.get('step', '') == self.deploy_group:
                # Use the specific notify_type if available else use slack_notify
                return step.get(notify_type, step.get('slack_notify', False))
        return False

    def _notify_with_thread(self, notify_type, channels, initial_message, initial_blocks=None, followups=[]):
        """Start a new thread with `initial_message`, then add any followups.

        This helps keep extra detail just a click away, without disrupting the
        main channel with large descriptive messages.
        """
        if self.deploy_group_is_set_to_notify(notify_type):
            for channel in channels:
                resp = self.sc.post(channels=[channel], message=initial_message, blocks=initial_blocks)[0]

                # If we got an error from Slack, fall back to posting it without a thread.
                thread_ts = resp['message']['ts'] if resp and resp['ok'] else None

                for followup in followups:
                    self.sc.post(channels=[channel], message=followup, thread_ts=thread_ts)
        else:
            log.debug(f"{self.deploy_group}.{notify_type} isn't set to notify slack")

    def notify_after_mark(self, ret):
        if ret == 0:
            if self.old_commit is not None and self.commit != self.old_commit:
                message = (
                    f"*{self.service}* - Marked *{self.commit[:12]}* for deployment on *{self.deploy_group}*.\n"
                    f"{self.authors}"
                )
                blocks = None
                self._notify_with_thread(
                    notify_type='notify_after_mark',
                    channels=self.channels,
                    initial_message=message,
                    initial_blocks=blocks,
                    followups=[(
                        f"{self.url_message}\n"
                        "\n"
                        "Roll it back at any time with:\n"
                        f"`paasta rollback --service {self.service} --deploy-group {self.deploy_group} "
                        f"--commit {self.old_commit}`"
                    )],
                )
        else:
            if self.old_commit is not None and self.commit != self.old_commit:
                self._notify_with_thread(
                    notify_type='notify_after_mark',
                    channels=self.channels,
                    initial_message=(
                        f"*{self.service}* - mark-for-deployment failed on *{self.deploy_group}* for *{self.commit[:12]}*.\n"  # noqa: E501
                        f"{self.authors}"
                    ),
                    followups=[self.url_message],
                )

    def notify_after_good_deploy(self):
        if self.old_commit is not None and self.commit != self.old_commit:
            self._notify_with_thread(
                notify_type='notify_after_good_deploy',
                channels=self.channels,
                initial_message=(
                    f"*{self.service}* - Finished deployment of *{self.commit[:12]}* on *{self.deploy_group}*.\n"
                    f"{self.authors}"
                ),
                followups=[(
                    f"{self.url_message}\n"
                    "If you need to roll back, run:\n"
                    f"`paasta rollback --service {self.service} --deploy-group {self.deploy_group} "
                    f"--commit {self.old_commit}`"
                )],
            )

    def notify_after_auto_rollback(self):
        if self.old_commit is not None and self.commit != self.old_commit:
            message = (
                f"*{self.service}* - Deployment of {self.commit} for {self.deploy_group} *failed*!\n"
                f"Auto-rolling back to {self.old_commit}\n"
                f"{self.url_message}\n"
            )
            self._notify_with_thread(
                notify_type='notify_after_auto_rollback',
                channels=self.channels,
                initial_message=message,
            )

    def notify_after_abort(self):
        if self.old_commit is not None and self.commit != self.old_commit:
            self._notify_with_thread(
                notify_type='notify_after_abort',
                channels=self.channels,
                initial_message=(
                    f"*{self.service}* - Deployment of {self.commit[:12]} to {self.deploy_group} *aborted*, "
                    "but still marked for deployment. PaaSTA will keep trying to deploy it until it is healthy.\n"
                    f"{self.authors}"
                ),
                followups=[(
                    f"{self.url_message}\n"
                    "If you need to roll back, run:\n"
                    f"`paasta rollback --service {self.service} --deploy-group {self.deploy_group} --commit {self.commit}`"  # noqa: E501
                )],
            )


def get_deploy_info(service, soa_dir):
    file_path = os.path.join(soa_dir, service, 'deploy.yaml')
    return read_deploy(file_path)


def print_rollback_cmd(old_git_sha, commit, auto_rollback, service, deploy_group):
    if old_git_sha is not None and old_git_sha != commit and not auto_rollback:
        paasta_print()
        paasta_print("If you wish to roll back, you can run:")
        paasta_print()
        paasta_print(
            PaastaColors.bold("    paasta rollback --service {} --deploy-group {} --commit {} ".format(
                service, deploy_group, old_git_sha,
            )),
        )


def paasta_mark_for_deployment(args):
    """Wrapping mark_for_deployment"""
    if args.verbose:
        log.setLevel(level=logging.DEBUG)
    else:
        log.setLevel(level=logging.INFO)

    service = args.service
    if service and service.startswith('services-'):
        service = service.split('services-', 1)[1]
    validate_service_name(service, soa_dir=args.soa_dir)

    deploy_group = args.deploy_group
    in_use_deploy_groups = list_deploy_groups(
        service=service,
        soa_dir=args.soa_dir,
    )
    _, invalid_deploy_groups = validate_given_deploy_groups(in_use_deploy_groups, [deploy_group])

    if len(invalid_deploy_groups) == 1:
        paasta_print(PaastaColors.red(
            "ERROR: These deploy groups are not currently used anywhere: %s.\n" % (",").join(invalid_deploy_groups),
        ))
        paasta_print(PaastaColors.red(
            "This isn't technically wrong because you can mark-for-deployment before deploying there",
        ))
        paasta_print(PaastaColors.red("but this is probably a typo. Did you mean one of these in-use deploy groups?:"))
        paasta_print(PaastaColors.red("   %s" % (",").join(in_use_deploy_groups)))
        paasta_print()
        paasta_print(PaastaColors.red("Continuing regardless..."))

    if args.git_url is None:
        args.git_url = get_git_url(service=service, soa_dir=args.soa_dir)

    commit = validate_git_sha(sha=args.commit, git_url=args.git_url)

    old_git_sha = get_currently_deployed_sha(service=service, deploy_group=deploy_group)
    if old_git_sha == commit:
        paasta_print("Warning: The sha asked to be deployed already matches what is set to be deployed:")
        paasta_print(old_git_sha)
        paasta_print("Continuing anyway.")

    if args.verify_image:
        if not is_docker_image_already_in_registry(service, args.soa_dir, commit):
            raise ValueError('Failed to find image in the registry for the following sha %s' % commit)

    deploy_info = get_deploy_info(service=service, soa_dir=args.soa_dir)

    if args.auto_rollback or args.block:
        deploy_process = MarkForDeploymentProcess(
            service=service,
            deploy_info=deploy_info,
            deploy_group=deploy_group,
            commit=commit,
            old_git_sha=old_git_sha,
            git_url=args.git_url,
            auto_rollback=args.auto_rollback,
            block=args.block,
            soa_dir=args.soa_dir,
            timeout=args.timeout,
            auto_certify_delay=args.auto_certify_delay,
            auto_abandon_delay=args.auto_abandon_delay,
            auto_rollback_delay=args.auto_rollback_delay,
        )
        ret = deploy_process.run()
        return ret
    else:
        # TODO: delete this branch once the state machine version is well tested. It should be equivalent.
        slack_notifier = SlackDeployNotifier(
            deploy_info=deploy_info, service=service,
            deploy_group=deploy_group, commit=commit, old_commit=old_git_sha, git_url=args.git_url,
            auto_rollback=args.auto_rollback,
        )

        ret = mark_for_deployment(
            git_url=args.git_url,
            deploy_group=deploy_group,
            service=service,
            commit=commit,
        )
        slack_notifier.notify_after_mark(ret=ret)
        print_rollback_cmd(old_git_sha, commit, args.auto_rollback, service, deploy_group)
        if args.block and ret == 0:
            try:
                wait_for_deployment(
                    service=service,
                    deploy_group=deploy_group,
                    git_sha=commit,
                    soa_dir=args.soa_dir,
                    timeout=args.timeout,
                )
                line = f"Deployment of {commit} for {deploy_group} complete"
                _log(
                    service=service,
                    component='deploy',
                    line=line,
                    level='event',
                )
                slack_notifier.notify_after_good_deploy()
            except (KeyboardInterrupt, TimeoutError):
                if args.auto_rollback is True:
                    if old_git_sha == commit:
                        paasta_print("Error: --auto-rollback was requested, but the previous sha")
                        paasta_print("is the same that was requested with --commit. Can't rollback")
                        paasta_print("automatically.")
                    else:
                        paasta_print("Auto-Rollback requested. Marking the previous sha")
                        paasta_print(f"({deploy_group}) for {old_git_sha} as desired.")
                        mark_for_deployment(
                            git_url=args.git_url,
                            deploy_group=deploy_group,
                            service=service,
                            commit=old_git_sha,
                        )
                        slack_notifier.notify_after_auto_rollback()
                else:
                    report_waiting_aborted(service, deploy_group)
                    slack_notifier.notify_after_abort()
                ret = 1
            except NoSuchCluster:
                report_waiting_aborted(service, deploy_group)
                slack_notifier.notify_after_abort()
        print_rollback_cmd(old_git_sha, commit, args.auto_rollback, service, deploy_group)
        return ret


class Progress():
    def __init__(self, percent=0, waiting_on=None, eta=None):
        self.percent = percent
        self.waiting_on = waiting_on

    def human_readable(self):
        if self.percent != 0 and self.percent != 100:
            s = f"{round(self.percent)}% (Waiting on {self.human_waiting_on()})"
        else:
            s = f"{round(self.percent)}%"
        return s

    def human_waiting_on(self):
        if self.waiting_on is None:
            return "N/A"
        things = []
        for cluster, queue in self.waiting_on.items():
            queue_length = len(queue)
            if queue_length == 0:
                continue
            elif queue_length == 1:
                things.append(f"`{cluster}`: `{queue[0].get_instance()}`")
            else:
                things.append(f"`{cluster}`: {len(queue)} instances")
        return ", ".join(things)


class MarkForDeploymentProcess(SLOSlackDeploymentProcess):
    rollback_states = ['start_rollback', 'rolling_back', 'rolled_back']
    rollforward_states = ['start_deploy', 'deploying', 'deployed']

    def __init__(
        self,
        service,
        deploy_info,
        deploy_group,
        commit,
        old_git_sha,
        git_url,
        auto_rollback,
        block,
        soa_dir,
        timeout,
        auto_certify_delay,
        auto_abandon_delay,
        auto_rollback_delay,
    ):
        self.service = service
        self.deploy_info = deploy_info
        self.deploy_group = deploy_group
        self.commit = commit
        self.old_git_sha = old_git_sha
        self.git_url = git_url
        self.auto_rollback = auto_rollback and old_git_sha is not None and old_git_sha != commit
        self.auto_rollbacks_ever_enabled = self.auto_rollback
        self.block = block
        self.soa_dir = soa_dir
        self.timeout = timeout
        self.mark_for_deployment_return_code = -1
        # Separate green_light per commit, so that we can tell wait_for_deployment for one commit to shut down
        # and quickly launch wait_for_deployment for another commit without causing a race condition.
        self.wait_for_deployment_green_lights = defaultdict(Event)

        self.human_readable_status = "Waiting on mark-for-deployment to initialize..."
        self.progress = Progress()
        self.slack_ts = None
        self.slack_channel = self.get_slack_channel()
        self.slack_channel_id = None
        self.last_action = None
        self.auto_certify_delay = auto_certify_delay
        self.auto_abandon_delay = auto_abandon_delay
        self.auto_rollback_delay = auto_rollback_delay
        self.slo_watchers = []

        self.start_slo_watcher_threads(self.service)
        # Initialize Slack threads and send the first message
        super().__init__()
        self.ping_authors()

    def get_progress(self) -> str:
        return self.progress.human_readable()

    def ping_authors(self):
        authors = get_authors_to_be_notified(git_url=self.git_url, from_sha=self.old_git_sha, to_sha=self.commit)
        self.update_slack_thread(authors)

    def get_slack_client(self) -> SlackClient:
        return get_slack_client().sc

    def get_slack_channel(self) -> str:
        """ Safely get some slack channel to post to. Defaults to ``DEFAULT_SLACK_CHANNEL``.
        Currently only uses the first slack channel available, and doesn't support
        multi-channel notifications. """
        if self.deploy_info.get('slack_notify', True):
            try:
                return self.deploy_info.get('slack_channels')[0]
            except (IndexError, AttributeError, TypeError):
                return DEFAULT_SLACK_CHANNEL
        else:
            return DEFAULT_SLACK_CHANNEL

    def get_deployment_name(self) -> str:
        return f"Deploy of `{self.commit[:8]}` of `{self.service}` to `{self.deploy_group}`:"

    def on_enter_start_deploy(self):
        self.update_slack_status(f"Marking `{self.commit[:8]}` for deployment for {self.deploy_group}...")
        self.mark_for_deployment_return_code = mark_for_deployment(
            git_url=self.git_url,
            deploy_group=self.deploy_group,
            service=self.service,
            commit=self.commit,
        )
        if self.mark_for_deployment_return_code != 0:
            self.trigger('mfd_failed')
        else:
            self.update_slack_thread(f"Marked `{self.commit[:8]}` for {self.deploy_group}.")
            log.debug("triggering mfd_succeeded")
            self.trigger('mfd_succeeded')

    def states(self) -> Collection[str]:
        return [
            "_begin",

            "start_deploy",
            "deploying",
            "deployed",

            "mfd_failed",
            "deploy_errored",
            "deploy_cancelled",

            "start_rollback",
            "rolling_back",
            "rolled_back",

            "abandon",
            "complete",
        ]

    def start_state(self) -> str:
        return "_begin"

    def start_transition(self) -> str:
        return "start_deploy"

    def valid_transitions(self) -> Iterator[state_machine.TransitionDefinition]:
        rollback_is_possible = (self.old_git_sha is not None and self.old_git_sha != self.commit)

        yield {
            'source': '_begin',
            'dest': 'start_deploy',
            'trigger': 'start_deploy',
        }
        yield {
            'source': 'start_deploy',
            'dest': 'deploying',
            'trigger': 'mfd_succeeded',
        }
        yield {
            'source': 'deploying',
            'dest': 'deployed',
            'trigger': 'deploy_finished',
        }

        yield {
            'source': ['start_deploy', 'start_rollback'],
            'dest': 'mfd_failed',
            'trigger': 'mfd_failed',
        }
        yield {
            'source': [s for s in self.states() if not self.is_terminal_state(s)],
            'dest': 'deploy_errored',
            'trigger': 'deploy_errored',
        }
        yield {

            'source': [s for s in self.states() if not self.is_terminal_state(s)],
            'dest': 'deploy_cancelled',
            'trigger': 'deploy_cancelled',
        }

        if rollback_is_possible:
            yield {
                'source': self.rollforward_states,
                'dest': 'start_rollback',
                'trigger': 'rollback_button_clicked',
            }
            yield {
                'source': self.rollback_states,
                'dest': None,  # this makes it an "internal transition", effectively a noop.
                'trigger': 'rollback_button_clicked',
            }
            yield {
                'source': self.rollforward_states,
                'dest': 'start_rollback',
                'trigger': 'rollback_slo_failure',
            }
            yield {
                'source': self.rollback_states,
                'dest': None,  # this makes it an "internal transition", effectively a noop.
                'trigger': 'rollback_slo_failure',
            }
            yield {
                'source': self.rollback_states,
                'dest': 'start_deploy',
                'trigger': 'forward_button_clicked',
            }
            yield {
                'source': self.rollforward_states,
                'dest': None,  # this makes it an "internal transition", effectively a noop.
                'trigger': 'forward_button_clicked',
            }
            yield {
                'source': 'start_rollback',
                'dest': 'rolling_back',
                'trigger': 'mfd_succeeded',
            }
            yield {
                'source': 'rolling_back',
                'dest': 'rolled_back',
                'trigger': 'deploy_finished',
            }

        yield {
            'source': 'deployed',
            'dest': 'complete',
            'trigger': 'complete_button_clicked',
        }
        yield {
            'source': 'deployed',
            'dest': 'complete',
            'trigger': 'auto_certify',
        }
        yield {
            'source': ['rolled_back', 'rolling_back'],
            'dest': 'abandon',
            'trigger': 'abandon_button_clicked',
        }
        yield {
            'source': 'rolled_back',
            'dest': 'abandon',
            'trigger': 'auto_abandon',
        }

        if rollback_is_possible:
            # Suppress these buttons if it doesn't make sense to roll back.
            yield {
                'source': '*',
                'dest': None,  # Don't actually change state, just call the before function.
                'trigger': 'enable_auto_rollbacks_button_clicked',
                'unless': [self.auto_rollbacks_enabled],
                'before': self.enable_auto_rollbacks,
            }
            yield {
                'source': '*',
                'dest': None,  # Don't actually change state, just call the before function.
                'trigger': 'disable_auto_rollbacks_button_clicked',
                'conditions': [self.any_slo_failing, self.auto_rollbacks_enabled],
                'before': self.disable_auto_rollbacks,
            }
        yield {
            'source': '*',
            'dest': None,
            'trigger': 'slos_started_failing',
            'conditions': [self.auto_rollbacks_enabled],
            'unless': [self.already_rolling_back],
            'before': self.start_auto_rollback_countdown,
        }
        yield {
            'source': '*',
            'dest': None,
            'trigger': 'slos_stopped_failing',
            'before': self.cancel_auto_rollback_countdown,
        }
        yield {
            'source': '*',
            'dest': None,
            'trigger': 'snooze_button_clicked',
            'before': self.restart_timer,
            'conditions': [self.is_timer_running],
        }

    def disable_auto_rollbacks(self):
        self.cancel_auto_rollback_countdown()
        self.auto_rollback = False

    def enable_auto_rollbacks(self):
        self.auto_rollback = True
        self.auto_rollbacks_ever_enabled = True

    def auto_rollbacks_enabled(self) -> bool:
        """This getter exists so it can be a condition on transitions, since those need to be callables."""
        return self.auto_rollback

    def get_auto_rollback_delay(self) -> float:
        return self.auto_rollback_delay

    def get_auto_certify_delay(self) -> float:
        if self.auto_certify_delay is not None:
            return self.auto_certify_delay
        else:
            if self.auto_rollbacks_ever_enabled:
                return DEFAULT_AUTO_CERTIFY_DELAY
            else:
                return 0

    def already_rolling_back(self) -> bool:
        return self.state in self.rollback_states

    def status_code_by_state(self) -> Mapping[str, int]:
        codes = {
            'deploy_errored': 2,
            'deploy_cancelled': 1,
            'mfd_failed': self.mark_for_deployment_return_code,
            'abandon': 1,
            'complete': 0,
        }

        if not self.block:
            # If we don't pass --wait-for-deployment, then exit immediately after mark-for-deployment succeeds.
            codes['deploying'] = 0
        if self.get_auto_certify_delay() <= 0:
            # Instead of setting a 0-second timer to move to certify, just exit 0 when the deploy finishes.
            codes['deployed'] = 0

        return codes

    def get_active_button(self) -> Optional[str]:
        return {
            'start_deploy': 'forward',
            'deploying': 'forward',
            'deployed': None,
            'start_rollback': 'rollback',
            'rolling_back': 'rollback',
            'rolled_back': None,
        }.get(self.state)

    def on_enter_mfd_failed(self):
        self.update_slack_status(f"Marking `{self.commit[:8]}` for deployment for {self.deploy_group} failed. Please see Jenkins for more output.")  # noqa E501

    def on_enter_deploying(self):
        # if self.block is False, then deploying is a terminal state so we will promptly exit.
        # Don't bother starting the background thread in this case.
        if self.block:
            thread = Thread(target=self.do_wait_for_deployment, args=(self.commit,), daemon=True)
            thread.start()

    def on_exit_deploying(self):
        self.wait_for_deployment_green_lights[self.commit].clear()

    def on_enter_start_rollback(self):
        self.update_slack_status(f"Rolling back ({self.deploy_group}) to {self.old_git_sha}")
        self.mark_for_deployment_return_code = mark_for_deployment(
            git_url=self.git_url,
            deploy_group=self.deploy_group,
            service=self.service,
            commit=self.old_git_sha,
        )

        if self.mark_for_deployment_return_code != 0:
            self.trigger('mfd_failed')
        else:
            self.update_slack_thread(f"Marked `{self.old_git_sha[:8]}` for {self.deploy_group}.")
            self.trigger('mfd_succeeded')

    def on_enter_rolling_back(self):
        if self.block:
            thread = Thread(target=self.do_wait_for_deployment, args=(self.old_git_sha,), daemon=True)
            thread.start()

    def on_exit_rolling_back(self):
        self.wait_for_deployment_green_lights[self.old_git_sha].clear()

    def on_enter_deploy_errored(self):
        report_waiting_aborted(self.service, self.deploy_group)
        self.update_slack_status(f"Deploy aborted, but it will still try to converge.")
        self.send_manual_rollback_instructions()

    def do_wait_for_deployment(self, target_commit: str):
        try:
            self.wait_for_deployment_green_lights[target_commit].set()
            wait_for_deployment(
                service=self.service,
                deploy_group=self.deploy_group,
                git_sha=target_commit,
                soa_dir=self.soa_dir,
                timeout=self.timeout,
                green_light=self.wait_for_deployment_green_lights[target_commit],
                progress=self.progress,
            )
            self.update_slack_thread(f"Finished waiting for deployment of {target_commit}")
            self.trigger('deploy_finished')

        except (KeyboardInterrupt, TimeoutError):
            if self.wait_for_deployment_green_lights[target_commit].is_set():
                # When we manually trigger a rollback, we clear the green_light, which causes wait_for_deployment to
                # raise KeyboardInterrupt. Don't trigger deploy_cancelled in this case.
                self.trigger('deploy_cancelled')
        except NoSuchCluster:
            self.trigger('deploy_errored')
        except Exception:
            log.error('Caught exception in wait_for_deployment:')
            log.error(traceback.format_exc())
            self.trigger('deploy_errored')

    def on_enter_rolled_back(self):
        self.update_slack_status(f"Finished rolling back to `{self.old_git_sha[:8]}` in {self.deploy_group}")
        line = f"Rollback to {self.old_git_sha[:8]} for {self.deploy_group} complete"
        _log(
            service=self.service,
            component='deploy',
            line=line,
            level='event',
        )
        self.start_timer(self.auto_abandon_delay, 'auto_abandon', 'abandon')

    def on_enter_deployed(self):
        self.update_slack_status(f"Finished deployment of `{self.commit[:8]}` to {self.deploy_group}")
        line = f"Deployment of {self.commit[:8]} for {self.deploy_group} complete"
        _log(
            service=self.service,
            component='deploy',
            line=line,
            level='event',
        )
        self.send_manual_rollback_instructions()
        if not (self.any_slo_failing() and self.auto_rollbacks_enabled()):
            if self.get_auto_certify_delay() > 0:
                self.start_timer(self.get_auto_certify_delay(), 'auto_certify', 'certify')

    def send_manual_rollback_instructions(self):
        if self.old_git_sha != self.commit:
            message = (
                "If you need to roll back manually, run: "
                f"`paasta rollback --service {self.service} --deploy-group {self.deploy_group} "
                f"--commit {self.old_git_sha}`"
            )
            self.update_slack_thread(message)
            paasta_print(message)

    def after_state_change(self):
        self.update_slack()
        super().after_state_change()

    def get_signalfx_api_token(self) -> str:
        return load_system_paasta_config().get_monitoring_config()['signalfx_api_key']

    def get_button_text(self, button, is_active) -> str:
        active_button_texts = {
            "forward": f"Rolling Forward to {self.commit[:8]} :zombocom:",
        }
        inactive_button_texts = {
            "forward": f"Continue Forward to {self.commit[:8]} :arrow_forward:",
            "complete": f"Complete deploy to {self.commit[:8]} :white_check_mark:",
            "snooze": f"Reset countdown",
            "enable_auto_rollbacks": "Enable auto rollbacks :eyes:",
            "disable_auto_rollbacks": "Disable auto rollbacks :close_eyes_monkey:",
        }

        if self.old_git_sha is not None:
            active_button_texts.update({
                "rollback": f"Rolling Back to {self.old_git_sha[:8]} :zombocom:",
            })
            inactive_button_texts.update({
                "rollback": f"Roll Back to {self.old_git_sha[:8]} :arrow_backward:",
                "abandon": f"Abandon deploy, staying on {self.old_git_sha[:8]} :x:",
            })

        return (active_button_texts if is_active else inactive_button_texts)[button]


class ClusterData:
    """An auxiliary data transfer class.

    Used by _query_clusters(), instances_deployed(),
    _run_cluster_worker(), _run_instance_worker().

    :param cluster: the name of the cluster.
    :param service: the name of the service.
    :param git_sha: git sha marked for deployment.
    :param instances_queue: a thread-safe queue. Should contain all cluster
                            instances that need to be checked.
    :type instances_queue: Queue
    """

    def __init__(self, cluster, service, git_sha, instances_queue):
        self.cluster = cluster
        self.service = service
        self.git_sha = git_sha
        self.instances_queue = instances_queue


def instances_deployed(cluster_data, instances_out, green_light):
    """Create a thread pool to run _run_instance_worker()

    :param cluster_data: an instance of ClusterData.
    :param instances_out: a empty thread-safe queue. I will contain
                          instances that are not deployed yet.
    :type instances_out: Queue
    :param green_light: See the docstring for _query_clusters().
    """
    num_threads = min(5, cluster_data.instances_queue.qsize())

    workers_launched = []
    for _ in range(num_threads):
        worker = Thread(
            target=_run_instance_worker,
            args=(cluster_data, instances_out, green_light),
        )
        worker.start()
        workers_launched.append(worker)

    for worker in workers_launched:
        worker.join()


def _run_instance_worker(cluster_data, instances_out, green_light):
    """Get instances from the instances_in queue and check them one by one.

    If an instance isn't deployed, add it to the instances_out queue
    to re-check it later.

    :param cluster_data: an instance of ClusterData.
    :param instances_out: See the docstring for instances_deployed().
    :param green_light: See the docstring for _query_clusters().
    """

    api = client.get_paasta_api_client(cluster=cluster_data.cluster)
    if not api:
        log.warning("Couldn't reach the PaaSTA api for {}! Assuming it is not "
                    "deployed there yet.".format(cluster_data.cluster))
        while not cluster_data.instances_queue.empty():
            try:
                instance_config = cluster_data.instances_queue.get(block=False)
            except Empty:
                return
            cluster_data.instances_queue.task_done()
            instances_out.put(instance_config)

    while not cluster_data.instances_queue.empty() and green_light.is_set():
        try:
            instance_config = cluster_data.instances_queue.get(block=False)
        except Empty:
            return

        instance = instance_config.get_instance()

        log.debug("Inspecting the deployment status of {}.{} on {}"
                  .format(cluster_data.service, instance, cluster_data.cluster))
        try:
            status = None
            status = api.service.status_instance(
                service=cluster_data.service,
                instance=instance,
            ).result()
        except HTTPError as e:
            if e.response.status_code == 404:
                log.warning("Can't get status for instance {}, service {} in "
                            "cluster {}. This is normally because it is a new "
                            "service that hasn't been deployed by PaaSTA yet"
                            .format(
                                instance, cluster_data.service,
                                cluster_data.cluster,
                            ))
            else:
                log.warning("Error getting service status from PaaSTA API for {}: {}"
                            "{}".format(
                                cluster_data.cluster, e.response.status_code,
                                e.response.text,
                            ))
        except ConnectionError as e:
            log.warning("Error getting service status from PaaSTA API for {}:"
                        "{}".format(cluster_data.cluster, e))

        long_running_status = None
        if status:
            if status.marathon:
                long_running_status = status.marathon
            elif status.kubernetes:
                long_running_status = status.kubernetes
        if not status:
            log.debug("No status for {}.{}, in {}. Not deployed yet."
                      .format(
                          cluster_data.service, instance,
                          cluster_data.cluster,
                      ))
            cluster_data.instances_queue.task_done()
            instances_out.put(instance_config)
        elif not long_running_status:
            log.debug("{}.{} in {} is not a Marathon or Kubernetes job. Marked as deployed."
                      .format(
                          cluster_data.service, instance,
                          cluster_data.cluster,
                      ))
        elif (
            long_running_status.expected_instance_count == 0
            or long_running_status.desired_state == 'stop'
        ):
            log.debug("{}.{} in {} is marked as stopped. Marked as deployed."
                      .format(
                          cluster_data.service, status.instance,
                          cluster_data.cluster,
                      ))
        else:
            if long_running_status.app_count != 1:
                paasta_print("  {}.{} on {} is still bouncing, {} versions "
                             "running"
                             .format(
                                 cluster_data.service, status.instance,
                                 cluster_data.cluster,
                                 long_running_status.app_count,
                             ))
                cluster_data.instances_queue.task_done()
                instances_out.put(instance_config)
                continue
            if not cluster_data.git_sha.startswith(status.git_sha):
                paasta_print("  {}.{} on {} doesn't have the right sha yet: {}"
                             .format(
                                 cluster_data.service, instance,
                                 cluster_data.cluster, status.git_sha,
                             ))
                cluster_data.instances_queue.task_done()
                instances_out.put(instance_config)
                continue
            if long_running_status.deploy_status not in ['Running', 'Deploying', 'Waiting']:
                paasta_print("  {}.{} on {} isn't running yet: {}"
                             .format(
                                 cluster_data.service, instance,
                                 cluster_data.cluster,
                                 long_running_status.deploy_status,
                             ))
                cluster_data.instances_queue.task_done()
                instances_out.put(instance_config)
                continue

            # The bounce margin factor defines what proportion of instances we need to be "safe",
            # so consider it scaled up "enough" if we have that proportion of instances ready.
            required_instance_count = int(math.ceil(
                instance_config.get_bounce_margin_factor() * long_running_status.expected_instance_count,
            ))
            if required_instance_count > long_running_status.running_instance_count:
                paasta_print("  {}.{} on {} isn't scaled up yet, "
                             "has {} out of {} required instances (out of a total of {})"
                             .format(
                                 cluster_data.service, instance,
                                 cluster_data.cluster,
                                 long_running_status.running_instance_count,
                                 required_instance_count,
                                 long_running_status.expected_instance_count,
                             ))
                cluster_data.instances_queue.task_done()
                instances_out.put(instance_config)
                continue
            paasta_print("Complete: {}.{} on {} looks 100% deployed at {} "
                         "instances on {}"
                         .format(
                             cluster_data.service, instance,
                             cluster_data.cluster,
                             long_running_status.running_instance_count,
                             status.git_sha,
                         ))
            cluster_data.instances_queue.task_done()


def _query_clusters(clusters_data, green_light):
    """Run _run_cluster_worker() in a separate thread for each paasta cluster

    :param clusters_data: a list of ClusterData instances.
    :param green_light: an instance of threading.Event().
                        It is supposed to be cleared when KeyboardInterrupt is
                        received. All running threads should check it
                        periodically and exit when it is cleared.
    """
    workers_launched = []

    for cluster_data in clusters_data:
        if not cluster_data.instances_queue.empty():
            worker = Thread(
                target=_run_cluster_worker,
                args=(cluster_data, green_light),
            )
            worker.start()
            workers_launched.append(worker)

    for worker in workers_launched:
        try:
            while green_light.is_set() and worker.isAlive():
                time.sleep(.2)
        except (KeyboardInterrupt, SystemExit):
            green_light.clear()
            paasta_print('KeyboardInterrupt received. Terminating..')
        worker.join()


def _run_cluster_worker(cluster_data, green_light):
    """Run instances_deployed() for a cluster

    :param cluster_data: an instance of ClusterData.
    :param green_light: See the docstring for _query_clusters().
    """
    instances_out = Queue()
    instances_deployed(
        cluster_data=cluster_data,
        instances_out=instances_out,
        green_light=green_light,
    )
    cluster_data.instances_queue = instances_out
    if cluster_data.instances_queue.empty():
        paasta_print(f"Deploy to {cluster_data.cluster} complete!")


def wait_for_deployment(service, deploy_group, git_sha, soa_dir, timeout, green_light=None, progress=None):
    # Currently only 'marathon' instances are supported for wait_for_deployment because they
    # are the only thing that are worth waiting on.
    service_configs = PaastaServiceConfigLoader(service=service, soa_dir=soa_dir, load_deployments=False)

    total_instances = 0
    clusters_data = []
    api_endpoints = load_system_paasta_config().get_api_endpoints()
    for cluster in service_configs.clusters:
        if cluster not in api_endpoints:
            paasta_print(PaastaColors.red(
                'Cluster %s is NOT in paasta-api endpoints config.' %
                cluster,
            ))
            raise NoSuchCluster

        instances_queue = Queue()
        for instance_config in service_configs.instance_configs(
            cluster=cluster,
            instance_type_class=MarathonServiceConfig,
        ):
            if instance_config.get_deploy_group() == deploy_group:
                instances_queue.put(instance_config)
                total_instances += 1
        for instance_config in service_configs.instance_configs(
            cluster=cluster,
            instance_type_class=KubernetesDeploymentConfig,
        ):
            if instance_config.get_deploy_group() == deploy_group:
                instances_queue.put(instance_config)
                total_instances += 1

        if not instances_queue.empty():
            clusters_data.append(ClusterData(
                cluster=cluster,
                service=service,
                git_sha=git_sha,
                instances_queue=instances_queue,
            ))

    if not clusters_data:
        _log(
            service=service,
            component='deploy',
            line=("Couldn't find any marathon instances for service {} in deploy group {}. Exiting."
                  .format(service, deploy_group)),
            level='event',
        )
        return

    paasta_print("Waiting for deployment of {} for '{}' to complete..."
                 .format(git_sha, deploy_group))

    deadline = time.time() + timeout
    if green_light is None:
        green_light = Event()
    green_light.set()

    with progressbar.ProgressBar(maxval=total_instances) as bar:
        while time.time() < deadline:
            _query_clusters(clusters_data, green_light)
            if not green_light.is_set():
                raise KeyboardInterrupt

            finished_instances = total_instances - sum((
                c.instances_queue.qsize()
                for c in clusters_data
            ))
            bar.update(finished_instances)
            if progress is not None:
                progress.percent = bar.percentage
                progress.waiting_on = {c.cluster: list(c.instances_queue.queue) for c in clusters_data}

            if all((
                cluster.instances_queue.empty()
                for cluster in clusters_data
            )):
                sys.stdout.flush()
                if progress is not None:
                    progress.percent = 100.0
                    progress.waiting_on = None
                return 0
            else:
                time.sleep(min(60, timeout))
            sys.stdout.flush()

    _log(
        service=service,
        component='deploy',
        line=compose_timeout_message(clusters_data, timeout, deploy_group, service, git_sha),
        level='event',
    )
    raise TimeoutError


def compose_timeout_message(clusters_data, timeout, deploy_group, service, git_sha):
    cluster_instances = {}
    for c_d in clusters_data:
        while c_d.instances_queue.qsize() > 0:
            cluster_instances.setdefault(
                c_d.cluster,
                [],
            ).append(c_d.instances_queue.get(block=False).get_instance())
            c_d.instances_queue.task_done()

    paasta_status = []
    paasta_logs = []
    for cluster, instances in sorted(cluster_instances.items()):
        if instances:
            joined_instances = ','.join(instances)
            paasta_status.append('paasta status -c {cluster} -s {service} -i {instances}'
                                 .format(
                                     cluster=cluster, service=service,
                                     instances=joined_instances,
                                 ))
            paasta_logs.append('paasta logs -c {cluster} -s {service} -i {instances} -C deploy -l 1000'
                               .format(
                                   cluster=cluster, service=service,
                                   instances=joined_instances,
                               ))

    return ("\n\nTimed out after {timeout} seconds, waiting for {service} "
            "in {deploy_group} to be deployed by PaaSTA.\n"
            "This probably means the deploy hasn't succeeded. The new service "
            "might not be healthy or one or more clusters could be having issues.\n\n"
            "To debug try running:\n\n"
            "  {status_commands}\n\n  {logs_commands}"
            "\n\nIf the service is known to be slow to start you may wish to "
            "increase the timeout on this step.\n"
            "To wait a little longer run:\n\n"
            "  paasta wait-for-deployment -s {service} -l {deploy_group} -c {git_sha}"
            .format(
                timeout=timeout,
                deploy_group=deploy_group,
                service=service,
                git_sha=git_sha,
                status_commands='\n  '.join(paasta_status),
                logs_commands='\n  '.join(paasta_logs),
            ))


class NoSuchCluster(Exception):
    """To be raised by wait_for_deployment() when a service has a marathon config for
    a cluster that is not listed in /etc/paasta/api_endpoints.json.
    """
    pass



#!/usr/bin/env python
# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./paasta_dump_locally_running_services.py

Outputs a JSON-encoded list of services that are running on this host along
with the host port that each service is listening on.
"""
import json
import sys

from paasta_tools.marathon_tools import get_marathon_services_running_here_for_nerve
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import paasta_print


def main():
    local_services = get_marathon_services_running_here_for_nerve(
        cluster=None,
        soa_dir=DEFAULT_SOA_DIR,
    )
    paasta_print(json.dumps(local_services))
    sys.exit(0)


if __name__ == '__main__':
    main()
